/**
 * Copyright schukai GmbH and contributors 2023. All Rights Reserved.
 * Node module: @schukai/component-form
 * This file is licensed under the AGPLv3 License.
 * License text available at https://www.gnu.org/licenses/agpl-3.0.en.html
 */
import {instanceSymbol} from "@schukai/monster/source/constants.mjs";
import {Pathfinder} from "@schukai/monster/source/data/pathfinder.mjs";
import {
    addAttributeToken,
    removeAttributeToken
} from "@schukai/monster/source/dom/attributes.mjs";
import {ATTRIBUTE_ROLE} from "@schukai/monster/source/dom/constants.mjs";
import {
    assembleMethodSymbol,
    getSlottedElements,
    registerCustomElement,
} from "@schukai/monster/source/dom/customelement.mjs";
import {CustomElement, attributeObserverSymbol} from "@schukai/monster/source/dom/customelement.mjs";
import {findTargetElementFromEvent} from "@schukai/monster/source/dom/events.mjs";
import {getDocument} from "@schukai/monster/source/dom/util.mjs";
import {getGlobal} from "@schukai/monster/source/types/global.mjs";
import {ID} from "@schukai/monster/source/types/id.mjs";
import {DeadMansSwitch} from "@schukai/monster/source/util/deadmansswitch.mjs";
import {Processing} from "@schukai/monster/source/util/processing.mjs";
import {
    STYLE_DISPLAY_MODE_BLOCK,
} from "./constants.mjs";
import {ButtonBarStyleSheet} from "./stylesheet/button-bar.mjs";
import {positionPopper} from "./util/floating-ui.mjs";
import {convertToPixels} from "@schukai/monster/source/dom/dimension.mjs";

export {ButtonBar};

/**
 * @private
 * @type {symbol}
 */
const timerCallbackSymbol = Symbol("timerCallback");

/**
 * local symbol
 * @private
 * @type {symbol}
 */
const resizeObserverSymbol = Symbol("windowResizeObserver");

/**
 * @private
 * @type {symbol}
 */
const dimensionsSymbol = Symbol("dimensions");

/**
 * @private
 * @type {symbol}
 */
const controlElementSymbol = Symbol("controlElement");

/**
 * @private
 * @type {symbol}
 */
const buttonBarSlotElementSymbol = Symbol("buttonBarSlotElement");

/**
 * @private
 * @type {symbol}
 */
const popperSlotElementSymbol = Symbol("popperSlotElement");


/**
 * @private
 * @type {symbol}
 */
const buttonBarElementSymbol = Symbol("buttonBarElement");

/**
 * @private
 * @type {symbol}
 */
const popperElementSymbol = Symbol("popperElement");


/**
 * local symbol
 * @private
 * @type {symbol}
 */
const closeEventHandler = Symbol("closeEventHandler");

/**
 * @private
 * @type {symbol}
 */
const popperSwitchEventHandler = Symbol("popperSwitchEventHandler");

/**
 * @private
 * @type {symbol}
 */
const popperNavElementSymbol = Symbol("popperNavElement");

/**
 * @private
 * @type {symbol}
 */
const switchElementSymbol = Symbol("switchElement");


/**
 * @private
 * @type {string}
 */
const ATTRIBUTE_POPPER_POSITION = "data-monster-popper-position";

/**
 * This CustomControl creates a button bar.
 *
 * <img src="./images/button-bar.png">
 *
 * Dependencies: the system uses functions of the [monsterjs](https://monsterjs.org/) library.
 *
 * You can create this control either by specifying the HTML tag `<monster-tabs />` directly in the HTML or using
 * Javascript via the `document.createElement('monster-tabs');` method.
 *
 * ```html
 * <monster-tabs></monster-tabs>
 * ```
 *
 * Or you can create this CustomControl directly in Javascript:
 *
 * ```js
 * import {ButtonBar} from '@schukai/component-form/source/tab.js';
 * document.createElement('monster-tabs');
 * ```
 *
 * @startuml tabs.png
 * skinparam monochrome true
 * skinparam shadowing false
 * HTMLElement <|-- CustomElement
 * CustomElement <|-- ButtonBar
 * @enduml
 *
 * @since 3.19.0
 * @copyright schukai GmbH
 * @memberOf Monster.Components.Form
 * @summary A configurable tab control
 * @fires Monster.Components.Form.event:monster-fetched
 */
class ButtonBar extends CustomElement {
    /**
     * This method is called by the `instanceof` operator.
     * @returns {symbol}
     */
    static get [instanceSymbol]() {
        return Symbol.for("@schukai/component-form/button-bar@@instance");
    }

    /**
     * To set the options via the html tag the attribute `data-monster-options` must be used.
     * @see {@link https://monsterjs.org/en/doc/#configurate-a-monster-control}
     *
     * The individual configuration values can be found in the table.
     *
     * @property {Object} templates Template definitions
     * @property {string} templates.main Main template
     * @property {Object} labels
     * @property {Object} popper FloatingUI popper configuration
     * @property {string} popper.placement=top Placement of the popper
     * @property {number} popper.middleware [middlerware](https://floating-ui.js.org/docs/middleware) to be used
     */
    get defaults() {
        const obj = Object.assign({}, super.defaults, {
            templates: {
                main: getTemplate(),
            },
            labels: {},
            popper: {
                placement: "bottom",
            },
        });

        initDefaultsFromAttributes.call(this, obj);

        return obj

    }

    /**
     * This method is called internal and should not be called directly.
     */
    [assembleMethodSymbol]() {
        const self = this;
        super[assembleMethodSymbol]();

        self[dimensionsSymbol] = new Pathfinder({data: {}});

        initControlReferences.call(self);
        initEventhandler.call(self);

        // setup structure
        initButtonBar.call(self).then(() => {
            initPopperSwitch.call(self);
        });

    }

    /**
     * This method is called internal and should not be called directly.
     *
     * @return {CSSStyleSheet[]}
     */
    static getCSSStyleSheet() {
        return [ButtonBarStyleSheet];
    }

    /**
     * This method is called internal and should not be called directly.
     *
     * @return {string}
     */
    static getTag() {
        return "monster-button-bar";
    }

    /**
     * This method is called by the dom and should not be called directly.
     *
     * @return {void}
     */
    connectedCallback() {
        super.connectedCallback();

        const document = getDocument();
        const self = this;

        for (const [, type] of Object.entries(["click", "touch"])) {
            // close on outside ui-events
            document.addEventListener(type, this[closeEventHandler]);
        }

        setTimeout(() => {
            updatePopper.call(self);
            updateResizeObserverObservation.call(this);
        }, 0);


    }

    /**
     * This method determines which attributes are to be monitored by `attributeChangedCallback()`.
     *
     * @return {string[]}
     */
    static get observedAttributes() {
        let attributes = super.observedAttributes;
        attributes.push(ATTRIBUTE_POPPER_POSITION);
        return attributes;
    }

    /**
     * This method is called by the dom and should not be called directly.
     *
     * @return {void}
     */
    disconnectedCallback() {
        super.disconnectedCallback();

        const document = getDocument();

        // close on outside ui-events
        for (const [, type] of Object.entries(["click", "touch"])) {
            document.removeEventListener(type, this[closeEventHandler]);
        }

        disconnectResizeObserver.call(this);

    }

    /**
     * Close the slotted dialog.
     * @return {ButtonBar}
     */
    hideDialog() {
        hide.call(this);
        return this;
    }

    /**
     * Open the slotted dialog.
     * @return {ButtonBar}
     */
    showDialog() {
        show.call(this);
        return this;
    }

    /**
     * Toggle the slotted dialog.
     * @return {ButtonBar}
     */
    toggleDialog() {
        toggle.call(this);
        return this;
    }
}

/**
 * @private
 * @param obj
 * @returns {*}
 */
function initDefaultsFromAttributes(obj) {
    const self = this;

    if (self.hasAttribute(ATTRIBUTE_POPPER_POSITION)) {
        obj.popper.placement = self.getAttribute(ATTRIBUTE_POPPER_POSITION);
    }

    return obj;

}


/**
 * @private
 */
function initEventhandler() {
    const self = this;
    /**
     * @param {Event} event
     */
    self[closeEventHandler] = (event) => {
        const path = event.composedPath();

        for (const [, element] of Object.entries(path)) {
            if (element === self) {
                return;
            }
        }

        hide.call(self);
    };

    if (self[buttonBarSlotElementSymbol]) {
        self[buttonBarSlotElementSymbol].addEventListener("slotchange", (event) => {
            checkAndRearrangeButtons.call(self);
        });
    }

    if (self[popperElementSymbol]) {
        self[popperElementSymbol].addEventListener("slotchange", (event) => {
            checkAndRearrangeButtons.call(self);
        });
    }

    // data-monster-options
    self[attributeObserverSymbol][ATTRIBUTE_POPPER_POSITION] = function (value) {
        self.setOption("classes.button", value);
    };

    self[resizeObserverSymbol] = new ResizeObserver((entries) => {

        if (self[timerCallbackSymbol] instanceof DeadMansSwitch) {
            try {
                self[timerCallbackSymbol].touch();
                return;
            } catch (e) {
                // catch Error("has already run");
                if (e.message !== "has already run") {
                    throw e;
                }
                delete self[timerCallbackSymbol];
            }
        }

        self[timerCallbackSymbol] = new DeadMansSwitch(200, () => {
            setTimeout(() => {
                updatePopper.call(self);
                self[dimensionsSymbol].setVia("data.calculated", false)
                checkAndRearrangeButtons.call(self);
            }, 50);
        });
    });

    initSlotChangedHandler.call(self);

}

function initSlotChangedHandler() {
    const self = this;

    self[buttonBarElementSymbol].addEventListener('slotchange', () => {
        updateResizeObserverObservation.call(self);
    })
}

function checkAndRearrangeButtons() {
    const self = this;

    if (self[dimensionsSymbol].getVia("data.calculated", false) !== true) {
        calculateButtonBarDimensions.call(self);
    }

    rearrangeButtons.call(self);
}

/**
 * @private
 * @return {Object}
 */
function rearrangeButtons() {
    const self = this;

    const switchWidth = self[switchElementSymbol].offsetWidth;

    let sum = switchWidth;
    const space = self[dimensionsSymbol].getVia("data.space");

    const buttonReferences = self[dimensionsSymbol].getVia("data.buttonReferences");

    for (const ref of buttonReferences) {

        sum += self[dimensionsSymbol].getVia(`data.button.${ref}`);

        let elements = getSlottedElements.call(self, '[data-monster-reference="' + ref + '"]', null); // null ↦ o
        if (elements.size === 0) {
            elements = getSlottedElements.call(self, '[data-monster-reference="' + ref + '"]', "popper"); // null ↦ o
        }

        const nextValue = elements.values().next();
        if (!nextValue) {
            continue;
        }

        const element = nextValue?.value;
        if (!(element instanceof HTMLElement)) {
            continue;
        }
        
        const buttonLabel = element.innerText;

        if (sum > space) {
            element.setAttribute("slot", "popper");
        } else {
            element.removeAttribute("slot");
        }
    }

    const inVisibleButtons = getSlottedElements.call(self, ":scope", 'popper'); // null ↦ o
    if (inVisibleButtons.size > 0) {
        self[switchElementSymbol].classList.remove("hidden");
    } else {
        self[switchElementSymbol].classList.add("hidden");
        setTimeout(() => {
            hide.call(self);
        }, 1);

    }

}


/**
 * @private
 * @param {HTMLElement} node
 * @return {number}
 */
function calcBoxWidth(node) {
    let dim = getGlobal("window").getComputedStyle(node);
    let bounding = node.getBoundingClientRect();

    return (
        convertToPixels(dim["border-left-width"]) +
        convertToPixels(dim["padding-left"]) +
        convertToPixels(dim["margin-left"]) +
        bounding["width"] +
        convertToPixels(dim["border-right-width"]) +
        convertToPixels(dim["margin-right"]) +
        convertToPixels(dim["padding-left"])
    );
}

/**
 * @private
 * @return {Object}
 */
function calculateButtonBarDimensions() {
    const self = this;

    const computedStyle = getComputedStyle(self.parentElement)
    if (computedStyle === null) {
        throw new Error("no computed style");
    }

    let width = self.parentElement.clientWidth;
    if (computedStyle.getPropertyValue("box-sizing") !== "border-box") {
        width = computedStyle.getPropertyValue("width");
        self[dimensionsSymbol].setVia("data.space", convertToPixels(width));

    } else {
        const borderWidth = getComputedStyle(self).getPropertyValue("--monster-border-width");
        if (borderWidth === null) {
            throw new Error("border-width not defined");
        }

        const borderWidthWithoutUnit = convertToPixels(borderWidth);

        // space to be allocated 
        self[dimensionsSymbol].setVia("data.space", width - 2 * borderWidthWithoutUnit);

    }

    self[dimensionsSymbol].setVia("data.visible", !(width === 0));

    let buttonReferences = [];

    const visibleButtons = getSlottedElements.call(self, ":scope", null); // null ↦ o

    for (const [, button] of visibleButtons.entries()) {
        if (!button.hasAttribute("data-monster-reference")) {
            button.setAttribute("data-monster-reference", new ID("vb").toString());
        }

        const ref = button.getAttribute("data-monster-reference");
        if (ref === null) continue;

        buttonReferences.push(ref);
        self[dimensionsSymbol].setVia(`data.button.${ref}`, calcBoxWidth.call(self, button));
    }

    const invisibleButtons = getSlottedElements.call(self, ":scope", 'popper'); // null ↦ o
    for (const [, button] of invisibleButtons.entries()) {
        if (!button.hasAttribute("data-monster-reference")) {
            button.setAttribute("data-monster-reference", new ID("ib").toString());
        }

        const ref = button.getAttribute("data-monster-reference");
        if (ref === null) continue;

        if (ref.indexOf("ib") !== 0) {
            buttonReferences.push(ref);
        }

    }


    self[dimensionsSymbol].setVia("data.calculated", true);
    self[dimensionsSymbol].setVia("data.buttonReferences", buttonReferences);
}

/**
 * @private
 */
function updateResizeObserverObservation() {
    const self = this;

    self[resizeObserverSymbol].disconnect();

    const slottedNodes = getSlottedElements.call(self);
    slottedNodes.forEach((node) => {
        self[resizeObserverSymbol].observe(node);
    });

    if(self.parentElement === null) return;
    self[resizeObserverSymbol].observe(self.parentElement);

}

/**
 * @private
 */
function disconnectResizeObserver() {
    const self = this;
    if (self[resizeObserverSymbol] instanceof ResizeObserver) {
        self[resizeObserverSymbol].disconnect();
    }
}

/**
 * @private
 */
function toggle() {
    if (this[popperElementSymbol].style.display === STYLE_DISPLAY_MODE_BLOCK) {
        hide.call(this);
    } else {
        show.call(this);
    }
}

/**
 * @private
 */
function hide() {
    const self = this;

    self[popperElementSymbol].style.display = "none";
    removeAttributeToken(self[controlElementSymbol], "class", "open");

}

/**
 * @private
 * @this PopperButton
 */
function show() {
    const self = this;

    if (self.getOption("disabled", false) === true) {
        return;
    }

    if (self[popperElementSymbol].style.display === STYLE_DISPLAY_MODE_BLOCK) {
        return;
    }

    self[popperElementSymbol].style.visibility = "hidden";
    self[popperElementSymbol].style.display = STYLE_DISPLAY_MODE_BLOCK;

    addAttributeToken(self[controlElementSymbol], "class", "open");

    updatePopper.call(self);

}

/**
 * @private
 */
function updatePopper() {
    const self = this;

    if (self[popperElementSymbol].style.display !== STYLE_DISPLAY_MODE_BLOCK) {
        return;
    }

    if (self.getOption("disabled", false) === true) {
        return;
    }

    positionPopper.call(self, self[switchElementSymbol], self[popperElementSymbol], self.getOption("popper", {}));
}

/**
 * @private
 * @return {Select}
 * @throws {Error} no shadow-root is defined
 */
function initControlReferences() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    self[controlElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=control]`);
    self[buttonBarElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=button-bar]`);
    self[popperElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=popper]`);
    self[popperNavElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=popper-nav]`);
    self[switchElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=switch]`);

    self[buttonBarSlotElementSymbol] = null;
    if (self[buttonBarElementSymbol]) self[buttonBarSlotElementSymbol] = self[buttonBarElementSymbol].querySelector(`slot`);
    self[popperSlotElementSymbol] = null;
    if (self[popperElementSymbol]) self[popperSlotElementSymbol] = self[popperElementSymbol].querySelector(`slot`);

}


/**
 * @private
 * @return {Promise<unknown>}
 * @throws {Error} no shadow-root is defined
 *
 */
function initButtonBar() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    return new Processing(() => {
        checkAndRearrangeButtons.call(self);
    }).run();

}

/**
 * @private
 */
function initPopperSwitch() {
    const self = this;

    /**
     * @param {Event} event
     */
    self[popperSwitchEventHandler] = (event) => {
        const element = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "switch");

        if (element instanceof HTMLButtonElement) {
            toggle.call(self);
        }
    };

    for (const type of ["click", "touch"]) {
        self[switchElementSymbol].addEventListener(type, self[popperSwitchEventHandler]);
    }

}

/**
 * @private
 * @return {string}
 */
function getTemplate() {
    // language=HTML
    return `
        <div data-monster-role="control" part="control">
            <div data-monster-role="button-bar">
                <slot></slot>
                <div part="popper-nav" data-monster-role="popper-nav"
                     tabindex="-1">
                    <button part="popper-switch" data-monster-role="switch"
                            class="monster-button-outline-tertiary hidden">
                        <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor"
                             viewBox="0 0 16 16">
                            <path d="M9.5 13a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0zm0-5a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0zm0-5a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0z"/>
                        </svg>
                    </button>
                </div>
            </div>
            <div data-monster-role="popper">
                <slot name="popper"></slot>
            </div>
        </div>`;
}

registerCustomElement(ButtonBar);
