/**
 * Copyright schukai GmbH and contributors 2023. All Rights Reserved.
 * Node module: @schukai/component-form
 * This file is licensed under the AGPLv3 License.
 * License text available at https://www.gnu.org/licenses/agpl-3.0.en.html
 */
import {instanceSymbol} from "@schukai/monster/source/constants.mjs";
import {ATTRIBUTE_ROLE} from "@schukai/monster/source/dom/constants.mjs";
import {assembleMethodSymbol, registerCustomElement} from "@schukai/monster/source/dom/customelement.mjs";
import {typeOf} from "@schukai/monster/source/types/typeof.mjs";
import {Button} from "./button.mjs";
import {PopperButton} from "./popper-button.mjs";
import {ConfirmButtonStyleSheet} from "./stylesheet/confirm-button.mjs";
import {fireCustomEvent} from "@schukai/monster/source/dom/events.mjs";

import "./button.mjs";
import "./message-state-button.mjs";
import {PopperButtonStyleSheet} from "./stylesheet/popper-button.mjs";
import {StateButtonStyleSheet} from "./stylesheet/state-button.mjs";

export {ConfirmButton};

/**
 * @private
 * @type {symbol}
 */
const confirmButtonElementSymbol = Symbol("confirmButtonElement");

/**
 * @private
 * @type {symbol}
 */
const cancelButtonElementSymbol = Symbol("cancelButtonElement");

/**
 * This action callback executes the actions when a button is clicked.
 *
 * @callback Monster.Components.Form~exampleActionCallback
 * @param {Event} e Event
 * @memberOf Monster.Components.Form
 * @this {CustomControl}
 */

/**
 * This CustomControl creates a confirm button element with a variety of options.
 *
 * <img src="./images/confirm-button.png">
 *
 * Dependencies: the system uses functions of the [monsterjs](https://monsterjs.org/) library
 * as well as [pooperjs](https://popper.js.org/docs/v2/).
 *
 * You can create this control either by specifying the HTML tag <monster-confirm-button />` directly in the HTML or using
 * Javascript via the `document.createElement('monster-confirm-button');` method.
 *
 * ```html
 * <monster-confirm-button></monster-confirm-button>
 * ```
 *
 * Or you can create this CustomControl directly in Javascript:
 *
 * ```js
 * import {ConfirmButton} from '@schukai/component-form/source/confirm-button.js';
 * document.createElement('monster-confirm-button');
 * ```
 * The `data-monster-button-class` attribute can be used to change the CSS class of the button.
 *
 * @startuml confirm-button.png
 * skinparam monochrome true
 * skinparam shadowing false
 * HTMLElement <|-- CustomElement
 * CustomElement <|-- CustomControl
 * CustomControl <|-- Button
 * Button <|-- PopperButton
 * PopperButton <|-- ConfirmButton
 * @enduml
 *
 * @since 1.5.0
 * @copyright schukai GmbH
 * @memberOf Monster.Components.Form
 * @summary A confirm button
 */
class ConfirmButton extends PopperButton {
    /**
     * This method is called by the `instanceof` operator.
     * @returns {symbol}
     * @since 2.1.0
     */
    static get [instanceSymbol]() {
        return Symbol.for("@schukai/component-form/confirm-button@@instance");
    }

    /**
     * To set the options via the html tag the attribute `data-monster-options` must be used.
     * @see {@link https://monsterjs.org/en/doc/#configurate-a-monster-control}
     *
     * The individual configuration values can be found in the table.
     *
     * @property {Object} toggleEventType=click,touch List of event types to be observed for opening the confirm message
     * @property {Object} templates Template definitions
     * @property {string} templates.main Main template
     * @property {string} labels.button='<slot name="button"></slot>' Button label
     * @property {string} labels.message='<slot>Should I Stay or Should I Go?</slot>' Message
     * @property {string} labels.confirm='<slot name="confirm">Confirm</slot>' Confirm button label
     * @property {string} labels.cancel='<slot name="cancel">Cancel</slot>' Cancel button label
     * @property {Monster.Components.Form~exampleActionCallback} actions.confirm
     * @property {Monster.Components.Form~exampleActionCallback} actions.cancel List of event types to be observed for opening the dropdown
     * @property {Object} popper [PopperJS Options](https://popper.js.org/docs/v2/)
     * @property {string} popper.placement=top PopperJS placement
     * @extends {PopperButton}
     */
    get defaults() {
        const self = this;

        let obj = Object.assign({}, super.defaults, {
            templates: {
                main: getTemplate(),
            },
            labels: {
                button: '<slot name="button"></slot>',
                message: "<slot>Should I Stay or Should I Go?</slot>",
                confirm: '<slot name="confirm">Confirm</slot>',
                cancel: '<slot name="cancel">Cancel</slot>',
            },
        });

        obj['classes']['confirmButton'] = "monster-button-primary";
        obj['classes']['cancelButton'] = "monster-button-secondary";
        obj['actions']['cancel'] = (e) => {
            self.hideDialog();
        };

        obj['actions']['confirm'] = (e) => {
            throw new Error("this callback must be implemented");
        };

        return obj;

    }

    hideDialog() {
        const self = this;
        super.hideDialog();
        self[confirmButtonElementSymbol].removeState();
        return self;
    }

    /**
     *
     * @return {Monster.Components.Form.PopperButton}
     */
    [assembleMethodSymbol]() {
        super[assembleMethodSymbol]();
        const self = this;
        initControlReferences.call(self);
        initEventhandler.call(self);

    }

    /**
     * @return {string}
     */
    static getTag() {
        return "monster-confirm-button";
    }

    /**
     * @return {Array}
     */
    static getCSSStyleSheet() {
        let styles = PopperButton.getCSSStyleSheet();
        styles.push(ConfirmButtonStyleSheet);
        return styles;
    }

    /**
     *
     * @returns {Monster.Components.Form.PopperButton}
     */
    showDialog() {
        const self = this;
        const confirmButtonClass = self.getOption("classes.confirmButton");
        if (confirmButtonClass) {
            self[confirmButtonElementSymbol].setOption("classes.button", confirmButtonClass);
        }

        const cancelButtonClass = self.getOption("classes.cancelButton");
        if (cancelButtonClass) {
            self[cancelButtonElementSymbol].setOption("classes.button", cancelButtonClass);
        }

        return super.showDialog();
    }

    /**
     * Set the confirm button state
     *
     * @param {string} state
     * @param {number} timeout
     * @returns {Monster.Components.Form.ConfirmButton}
     */
    setConfirmButtonState(state, timeout) {
        const self = this;
        self[confirmButtonElementSymbol].setState(state, timeout);
        return self;
    }

    /**
     * Set and show the confirm button message
     *
     * @param {string} message
     * @param {number} timeout
     * @returns {Monster.Components.Form.ConfirmButton}
     */
    showConfirmButtonMessage(message, timeout) {
        const self = this;
        self[confirmButtonElementSymbol].setMessage(message);
        setTimeout(() => {
            self[confirmButtonElementSymbol].showMessage(timeout);
        }, 0);
        return self;
    }

}

/**
 * @private
 * @param type
 * @return {function}
 */
function getAction(type) {
    const self = this;

    return function (event) {
        const button = this;

        let callback = self.getOption(`actions.${type}`);
        if (typeOf(callback) !== "function") {
            return;
        }

        callback.call(button, event);
    };
}

/**
 * @private
 * @return {initEventhandler}
 * @fires Monster.Components.Form.PopperButton#monster-confirm
 * @fires Monster.Components.Form.PopperButton#monster-confirmed
 */
function initEventhandler() {
    const self = this;

    this[cancelButtonElementSymbol].setOption("clickEventType", self.getOption("clickEventType", ["click"]));
    this[cancelButtonElementSymbol].setOption("actions.click", getAction.call(self, "cancel"));

    this[confirmButtonElementSymbol].setOption("clickEventType", self.getOption("clickEventType", ["click"]));
    this[confirmButtonElementSymbol].setOption("actions.click", () => {
        const callback = getAction.call(self, "confirm");
        fireCustomEvent(self, "monster-confirm", {});
        callback.call(self);
        fireCustomEvent(self, "monster-confirmed", {});
    });

    return this;
}

/**
 * @private
 * @return {Select}
 */
function initControlReferences() {
    const self = this;
    self[confirmButtonElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=confirm]`);
    self[cancelButtonElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=cancel]`);
}

/**
 * @private
 * @return {string}
 */
function getTemplate() {
    // language=HTML
    return `

        <div data-monster-role="control" part="control">
            <button data-monster-attributes="disabled path:disabled | if:true, class path:classes.button"
                    data-monster-role="button"
                    part="button"
                    data-monster-replace="path:labels.button"></button>

            <div data-monster-role="popper" part="popper" tabindex="-1" class="monster-color-primary-1">
                <div data-monster-role="arrow"></div>
                <div part="message" class="flex" data-monster-replace="path:labels.message"></div>
                <div data-monster-role="decision">
                    <monster-message-state-button exportparts="monster-state-button:confirm-button,
                                          control:confirm-control,
                                          button-button:confirm-control-button,
                                          button-control:confirm-button-control,
                                          popper:confirm-popper,
                                          message:confirm-popper-message-message,
                                          icon:confirm-popper-message-icon,
                                          title:confirm-popper-message-title"
                                                  data-monster-attributes="data-monster-button-class path:classes.confirmButton"
                                                  data-monster-role="confirm" part="confirm">
                        <span data-monster-replace="path:labels.confirm"></span>
                    </monster-message-state-button>

                    <monster-button exportparts="button:cancel-button,control:cancel-control"
                                    data-monster-attributes="data-monster-button-class path:classes.cancelButton"
                                    data-monster-role="cancel" part="cancel">
                        <span data-monster-replace="path:labels.cancel"></span>
                    </monster-button>
                </div>
            </div>
        </div>
    `;
}

registerCustomElement(ConfirmButton);
