/**
 * Copyright schukai GmbH and contributors 2023. All Rights Reserved.
 * Node module: @schukai/component-form
 * This file is licensed under the AGPLv3 License.
 * License text available at https://www.gnu.org/licenses/agpl-3.0.en.html
 */
import { instanceSymbol } from "@schukai/monster/source/constants.mjs";
import {
    registerCustomElement,
} from "@schukai/monster/source/dom/customelement.mjs";
import { Reload } from "./reload.mjs";

export { ShadowReload };

/**
 * This CustomControl reloads the content of a url and embeds it into the dom.
 *
 * <img src="./images/shadowreload.png">
 *
 * You can create this control either by specifying the HTML tag `<monster-shadow-reload />` directly in the HTML or using
 * Javascript via the `document.createElement('monster-shadow-reload');` method.
 *
 * ```html
 * <monster-shadow-reload></monster-shadow-reload>
 *
 * <script type="module">
 * import {Reload} from '@schukai/component-form/source/shadow-reload.js';
 * document.createElement('monster-shadow-reload');
 * </script>
 * ```
 *
 * A simple configuration can look like this
 *
 * ```html
 * <script id="config"
 *         type="application/json">
 *     {
 *         "url": "./content.html",
 *     }
 * </script>
 *
 * <monster-reload data-monster-options-selector="#config">
 * </monster-reload>
 * ```
 *
 * If you want to display a loader, you can insert a div with the attribute `data-monster-role="container"`.
 * The content of this div will be replaced by the loaded code.
 *
 * ```html
 * <monster-reload data-monster-options-selector="#config">
 * <div data-monster-role="container">
 * LOADER ...
 * </div>
 * </monster-reload>
 * ```
 *
 * If you need additional structure, you can simply specify it.
 *
 * ```html
 * <monster-shadow-reload data-monster-options-selector="#config">
 * <div class="row">
 *   <div class="col" data-monster-role="container">
 *       LOADER ...
 *   </div>
 * </div>
 * </monster-shadow-reload>
 * ```
 *
 * @startuml shadowreload.png
 * skinparam monochrome true
 * skinparam shadowing false
 * HTMLElement <|-- CustomElement
 * CustomElement <|-- CustomControl
 * CustomControl <|-- Reload
 * Reload <|-- ShadowReload
 * @enduml
 *
 * @since 3.6.0
 * @copyright schukai GmbH
 * @memberOf Monster.Components.Form
 * @summary A shadow reload control
 * @fires Monster.Components.Form.event:monster-fetched
 */
class ShadowReload extends Reload {
    /**
     * This method is called by the `instanceof` operator.
     * @returns {symbol}
     */
    static get [instanceSymbol]() {
        return Symbol.for("@schukai/component-form/shadow-reload");
    }

    /**
     * To set the options via the html tag the attribute `data-monster-options` must be used.
     * @see {@link https://monsterjs.org/en/doc/#configurate-a-monster-control}
     *
     * The individual configuration values can be found in the table.
     *
     * @property {Object} templates Template definitions
     * @property {string} templates.main Main template
     * @property {string} url=undefined
     * @property {string} reload=undefined currently the values defined are `onshow` and `always`. The default `onshow` removes the IntersectionObserver. This means that the content is only loaded once. reloading of the content does not occur.
     * @property {string} filter=undefined dom selectors to search for elements, if undefined then everything is taken
     * @property {Monster.Components.Form.Processor[]} processors
     * @property {Object} fetch Fetch [see Using Fetch mozilla.org](https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch)
     * @property {String} fetch.redirect=error
     * @property {String} fetch.method=GET
     * @property {String} fetch.mode=same-origin
     * @property {String} fetch.credentials=same-origin
     * @property {Object} fetch.headers={"accept":"text/html"}}
     */
    get defaults() {
        return Object.assign(
            {},
            super.defaults,
            {
                templates: {
                    main: getTemplate.call(this),
                },
                shadowMode: "open"
            }
        );
    }


    /**
     * This method is called internal and should not be called directly.
     *
     * @return {string}
     */
    static getTag() {
        return "monster-shadow-reload";
    }

}

/**
 * @private
 * @return {string}
 */
function getTemplate() {
    return this.innerHTML;
}

registerCustomElement(ShadowReload);
