/**
 * Copyright schukai GmbH and contributors 2023. All Rights Reserved.
 * Node module: @schukai/component-form
 * This file is licensed under the AGPLv3 License.
 * License text available at https://www.gnu.org/licenses/agpl-3.0.en.html
 */
import { instanceSymbol } from "@schukai/monster/source/constants.mjs";
import { createPopper } from "@popperjs/core";
import { extend } from "@schukai/monster/source/data/extend.mjs";
import { Pathfinder } from "@schukai/monster/source/data/pathfinder.mjs";
import { addToObjectLink, hasObjectLink } from "@schukai/monster/source/dom/attributes.mjs";
import { ATTRIBUTE_PREFIX, ATTRIBUTE_ROLE } from "@schukai/monster/source/dom/constants.mjs";
import {
    assembleMethodSymbol,
    CustomElement,
    getSlottedElements,
    registerCustomElement,
} from "@schukai/monster/source/dom/customelement.mjs";
import { findTargetElementFromEvent } from "@schukai/monster/source/dom/events.mjs";
import { getDocument } from "@schukai/monster/source/dom/util.mjs";
import { random } from "@schukai/monster/source/math/random.mjs";
import { getGlobal } from "@schukai/monster/source/types/global.mjs";
import { ID } from "@schukai/monster/source/types/id.mjs";
import { isArray, isString } from "@schukai/monster/source/types/is.mjs";
import { TokenList } from "@schukai/monster/source/types/tokenlist.mjs";
import { clone } from "@schukai/monster/source/util/clone.mjs";
import { DeadMansSwitch } from "@schukai/monster/source/util/deadmansswitch.mjs";
import { Processing } from "@schukai/monster/source/util/processing.mjs";
import {
    ATTRIBUTE_BUTTON_LABEL,
    ATTRIBUTE_FORM_RELOAD,
    ATTRIBUTE_FORM_URL,
    STYLE_DISPLAY_MODE_BLOCK,
} from "./constants.mjs";

import { TabsStyleSheet } from "./stylesheet/tabs.mjs";
import { loadAndAssignContent } from "./util/fetch.mjs";
import { popperInstanceSymbol, setEventListenersModifiers } from "./util/popper.mjs";

export { Tabs };

/**
 * @private
 * @type {symbol}
 */
const popperElementSymbol = Symbol("popperElement");

/**
 * @private
 * @type {symbol}
 */
const popperNavElementSymbol = Symbol("popperNavElement");

/**
 * @private
 * @type {symbol}
 */
const controlElementSymbol = Symbol("controlElement");

/**
 * @private
 * @type {symbol}
 */
const navElementSymbol = Symbol("navElement");
/**
 * @private
 * @type {symbol}
 */
const switchElementSymbol = Symbol("switchElement");

/**
 * @private
 * @type {symbol}
 */
const changeTabEventHandler = Symbol("changeTabEventHandler");
/**
 * @private
 * @type {symbol}
 */
const removeTabEventHandler = Symbol("removeTabEventHandler");

/**
 * @private
 * @type {symbol}
 */
const popperSwitchEventHandler = Symbol("popperSwitchEventHandler");

/**
 * local symbol
 * @private
 * @type {symbol}
 */
const closeEventHandler = Symbol("closeEventHandler");

/**
 * @private
 * @type {symbol}
 */
const mutationObserverSymbol = Symbol("mutationObserver");

/**
 * @private
 * @type {symbol}
 */
const dimensionsSymbol = Symbol("dimensions");

/**
 * @private
 * @type {symbol}
 */
const timerCallbackSymbol = Symbol("timerCallback");

/**
 * local symbol
 * @private
 * @type {symbol}
 */
const resizeObserverSymbol = Symbol("resizeObserver");

/**
 * This CustomControl creates a tab element with a variety of options.
 *
 * <img src="./images/tabs.png">
 *
 * Dependencies: the system uses functions of the [monsterjs](https://monsterjs.org/) library.
 *
 * You can create this control either by specifying the HTML tag `<monster-tabs />` directly in the HTML or using
 * Javascript via the `document.createElement('monster-tabs');` method.
 *
 * ```html
 * <monster-tabs></monster-tabs>
 * ```
 *
 * Or you can create this CustomControl directly in Javascript:
 *
 * ```js
 * import {Tabs} from '@schukai/component-form/source/tab.js';
 * document.createElement('monster-tabs');
 * ```
 *
 * @startuml tabs.png
 * skinparam monochrome true
 * skinparam shadowing false
 * HTMLElement <|-- CustomElement
 * CustomElement <|-- CustomControl
 * CustomControl <|-- Tabs
 * @enduml
 *
 * @since 1.10.0
 * @copyright schukai GmbH
 * @memberOf Monster.Components.Form
 * @summary A configurable tab control
 * @fires Monster.Components.Form.event:monster-fetched
 */
class Tabs extends CustomElement {
    /**
     * This method is called by the `instanceof` operator.
     * @returns {symbol}
     * @since 2.1.0
     */
    static get [instanceSymbol]() {
        return Symbol.for("@schukai/component-form/tabs");
    }

    /**
     * To set the options via the html tag the attribute `data-monster-options` must be used.
     * @see {@link https://monsterjs.org/en/doc/#configurate-a-monster-control}
     *
     * The individual configuration values can be found in the table.
     *
     * @property {Object} templates Template definitions
     * @property {string} templates.main Main template
     * @property {Object} labels
     * @property {string} labels.new-tab-label="New Tab"
     * @property {Object} fetch Fetch [see Using Fetch mozilla.org](https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch)
     * @property {String} fetch.redirect=error
     * @property {String} fetch.method=GET
     * @property {String} fetch.mode=same-origin
     * @property {String} fetch.credentials=same-origin
     * @property {Object} fetch.headers={"accept":"text/html"}}
     * @property {Object} popper [PopperJS Options](https://popper.js.org/docs/v2/)
     * @property {string} popper.placement=bottom PopperJS placement
     * @property {Object[]} modifiers={name:offset} PopperJS placement
     */
    get defaults() {
        return Object.assign({}, super.defaults, {
            templates: {
                main: getTemplate(),
            },
            labels: {
                "new-tab-label": "New Tab",
            },
            buttons: {
                standard: [],
                popper: [],
            },
            fetch: {
                redirect: "error",
                method: "GET",
                mode: "same-origin",
                credentials: "same-origin",
                headers: {
                    accept: "text/html",
                },
            },
            popper: {
                placement: "bottom",
                modifiers: [
                    {
                        name: "offset",
                        options: {
                            offset: [0, 2],
                        },
                    },

                    {
                        name: "eventListeners",
                        enabled: false,
                    },
                ],
            },
        });
    }

    /**
     * This method is called internal and should not be called directly.
     */
    [assembleMethodSymbol]() {
        const self = this;
        super[assembleMethodSymbol]();

        initControlReferences.call(self);

        self[dimensionsSymbol] = new Pathfinder({ data: {} });
        // change and remove Tabs
        initEventhandler.call(self);

        // setup structure
        initTabButtons.call(self).then(() => {
            initPopperSwitch.call(self);
            initPopper.call(self);
            attachResizeObserver.call(self);
            attachTabChangeObserver.call(self);
        });

    }

    /**
     * This method is called internal and should not be called directly.
     *
     * @return {CSSStyleSheet[]}
     */
    static getCSSStyleSheet() {
        return [TabsStyleSheet];
    }

    /**
     * This method is called internal and should not be called directly.
     *
     * @return {string}
     */
    static getTag() {
        return "monster-tabs";
    }

    /**
     * This method is called by the dom and should not be called directly.
     *
     * @return {void}
     */
    connectedCallback() {
        super.connectedCallback();

        const document = getDocument();
        const self = this;

        for (const [, type] of Object.entries(["click", "touch"])) {
            // close on outside ui-events
            document.addEventListener(type, this[closeEventHandler]);
        }
    }

    /**
     * This method is called by the dom and should not be called directly.
     *
     * @return {void}
     */
    disconnectedCallback() {
        super.disconnectedCallback();

        const document = getDocument();

        // close on outside ui-events
        for (const [, type] of Object.entries(["click", "touch"])) {
            document.removeEventListener(type, this[closeEventHandler]);
        }
    }
}

/**
 * @private
 */
function initPopperSwitch() {
    const self = this;

    const nodes = getSlottedElements.call(self, `[${ATTRIBUTE_ROLE}="switch"]`); // null ↦ only unnamed slots
    let switchButton;
    if (nodes.size === 0) {
        switchButton = document.createElement("button");
        switchButton.setAttribute(ATTRIBUTE_ROLE, "switch");
        switchButton.setAttribute("part", "switch");
        switchButton.classList.add("hidden");
        switchButton.innerHTML =
            '<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" viewBox="0 0 16 16"><path d="M9.5 13a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0zm0-5a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0zm0-5a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0z"/></svg>';
        self[navElementSymbol].prepend(switchButton);
    } else {
        switchButton = nodes.next();
    }

    /**
     * @param {Event} event
     */
    self[popperSwitchEventHandler] = (event) => {
        const element = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "switch");

        if (element instanceof HTMLButtonElement) {
            togglePopper.call(self);
        }
    };

    for (const type of ["click", "touch"]) {
        switchButton.addEventListener(type, self[popperSwitchEventHandler]);
    }

    self[switchElementSymbol] = switchButton;
}

/**
 * @private
 */
function hidePopper() {
    const self = this;

    if (!self[popperInstanceSymbol]) {
        return;
    }

    self[popperElementSymbol].style.display = "none";
    // performance https://popper.js.org/docs/v2/tutorial/#performance
    setEventListenersModifiers.call(self, false);
}

/**
 * @private
 */
function showPopper() {
    const self = this;

    if (self[popperElementSymbol].style.display === STYLE_DISPLAY_MODE_BLOCK) {
        return;
    }

    self[popperElementSymbol].style.visibility = "hidden";
    self[popperElementSymbol].style.display = STYLE_DISPLAY_MODE_BLOCK;
    // performance https://popper.js.org/docs/v2/tutorial/#performance
    setEventListenersModifiers.call(self, true);

    self[popperInstanceSymbol].update();

    new Processing(() => {
        self[popperElementSymbol].style.removeProperty("visibility");
    }).run();
}

/**
 * @private
 */
function togglePopper() {
    const self = this;

    if (self[popperElementSymbol].style.display === STYLE_DISPLAY_MODE_BLOCK) {
        hidePopper.call(self);
    } else {
        showPopper.call(self);
    }
}

/**
 * @private
 */
function attachResizeObserver() {
    const self = this;

    // against flickering
    self[resizeObserverSymbol] = new ResizeObserver((entries) => {
        if (self[timerCallbackSymbol] instanceof DeadMansSwitch) {
            try {
                self[timerCallbackSymbol].touch();
                return;
            } catch (e) {
                delete self[timerCallbackSymbol];
            }
        }

        self[timerCallbackSymbol] = new DeadMansSwitch(200, () => {
            self[dimensionsSymbol].setVia("data.calculated", false);
            checkAndRearrangeButtons.call(self);
        });
    });

    self[resizeObserverSymbol].observe(self[navElementSymbol]);
}

/**
 * @private
 */
function attachTabChangeObserver() {
    const self = this;

    // against flickering
    new MutationObserver((mutations) => {
        let runUpdate = false;

        for (const mutation of mutations) {
            if (mutation.type === "childList") {
                if (mutation.addedNodes.length > 0 || mutation.removedNodes.length > 0) {
                    runUpdate = true;
                    break;
                }
            }
        }

        if (runUpdate === true) {
            self[dimensionsSymbol].setVia("data.calculated", false);
            initTabButtons.call(self);
        }
    }).observe(self, {
        childList: true,
    });
}

/**
 * @private
 * @return {Select}
 * @external "external:createPopper"
 * @see {@link  Plugins}
 */
function initPopper() {
    const self = this;

    let options = extend({}, self.getOption("popper"));

    self[popperInstanceSymbol] = createPopper(self[switchElementSymbol], self[popperElementSymbol], options);

    const observer1 = new MutationObserver(function (mutations) {
        let runUpdate = false;

        for (const mutation of mutations) {
            if (mutation.type === "childList") {
                if (mutation.addedNodes.length > 0 || mutation.removedNodes.length > 0) {
                    runUpdate = true;
                    break;
                }
            }
        }

        if (runUpdate === true) {
            self[popperInstanceSymbol].update();
        }
    });

    observer1.observe(self[popperNavElementSymbol], {
        childList: true,
        subtree: true,
    });

    return self;
}

/**
 * @private
 * @param {HTMLElement} element
 */
function show(element) {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    const reference = element.getAttribute(`${ATTRIBUTE_PREFIX}tab-reference`);

    const nodes = getSlottedElements.call(self);
    for (const node of nodes) {
        const id = node.getAttribute("id");
        if (id === reference) {
            node.classList.add("active");

            if (node.hasAttribute(ATTRIBUTE_FORM_URL)) {
                const url = node.getAttribute(ATTRIBUTE_FORM_URL);

                if (
                    !node.hasAttribute(ATTRIBUTE_FORM_RELOAD) ||
                    node.getAttribute(ATTRIBUTE_FORM_RELOAD).toLowerCase() === "onshow"
                ) {
                    node.removeAttribute(ATTRIBUTE_FORM_URL);
                }

                const options = self.getOption("fetch", {});
                loadAndAssignContent(node, url, options);
            }
        } else {
            node.classList.remove("active");
        }
    }

    const standardButtons = self.getOption("buttons.standard");
    for (const [index] in standardButtons) {
        const button = standardButtons[index];
        const state = button["reference"] === reference ? "active" : "inactive";
        self.setOption(`buttons.standard.${index}.state`, state);
    }

    const popperButton = self.getOption("buttons.popper");
    for (const [index] in popperButton) {
        const button = popperButton[index];
        const state = button["reference"] === reference ? "active" : "inactive";
        self.setOption(`buttons.popper.${index}.state`, state);
    }

    hidePopper.call(self);
}

/**
 * @private
 */
function initEventhandler() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    /**
     * @param {Event} event
     */
    self[changeTabEventHandler] = (event) => {
        const element = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "button");

        if (element instanceof HTMLButtonElement && element.disabled !== true) {
            show.call(self, element);
        }
    };

    /**
     * @param {Event} event
     */
    self[removeTabEventHandler] = (event) => {
        const element = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "remove-tab");

        if (element instanceof HTMLElement) {
            const button = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "button");

            if (button instanceof HTMLButtonElement && button.disabled !== true) {
                const reference = button.getAttribute(`${ATTRIBUTE_PREFIX}tab-reference`);
                if (reference) {
                    const container = self.querySelector(`[id=${reference}]`);
                    if (container instanceof HTMLElement) {
                        container.remove();
                        initTabButtons.call(self);
                    }
                }
            }
        }
    };

    self[navElementSymbol].addEventListener("touch", self[changeTabEventHandler]);
    self[navElementSymbol].addEventListener("click", self[changeTabEventHandler]);

    self[navElementSymbol].addEventListener("touch", self[removeTabEventHandler]);
    self[navElementSymbol].addEventListener("click", self[removeTabEventHandler]);

    /**
     * @param {Event} event
     */
    self[closeEventHandler] = (event) => {
        const path = event.composedPath();

        for (const [, element] of Object.entries(path)) {
            if (element === self) {
                return;
            }
        }

        hidePopper.call(self);
    };

    return self;
}

/**
 * @private
 * @param observedNode
 */
function attachTabMutationObserver(observedNode) {
    const self = this;

    if (hasObjectLink(observedNode, mutationObserverSymbol)) {
        return;
    }

    /**
     * this construct monitors a node whether it is disabled or modified
     * @type {MutationObserver}
     */
    const observer = new MutationObserver(function (mutations) {
        if (isArray(mutations)) {
            const mutation = mutations.pop();
            if (mutation instanceof MutationRecord) {
                initTabButtons.call(self);
            }
        }
    });

    observer.observe(observedNode, {
        childList: false,
        attributes: true,
        subtree: false,
        attributeFilter: ["disabled", ATTRIBUTE_BUTTON_LABEL, `${ATTRIBUTE_PREFIX}button-icon`],
    });

    addToObjectLink(observedNode, mutationObserverSymbol, observer);
}

/**
 * @private
 * @return {Select}
 * @throws {Error} no shadow-root is defined
 */
function initControlReferences() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    self[controlElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=control]`);
    self[navElementSymbol] = self.shadowRoot.querySelector(`nav[${ATTRIBUTE_ROLE}=nav]`);
    self[popperElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=popper]`);
    self[popperNavElementSymbol] = self.shadowRoot.querySelector(`[${ATTRIBUTE_ROLE}=popper-nav]`);
}

/**
 * @private
 * @return {Promise<unknown>}
 * @throws {Error} no shadow-root is defined
 *
 */
function initTabButtons() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    let activeReference;

    const dimensionsCalculated = self[dimensionsSymbol].getVia("data.calculated", false);

    const buttons = [];
    const nodes = getSlottedElements.call(self, undefined, null); // null ↦ only unnamed slots

    for (const node of nodes) {
        if (!(node instanceof HTMLElement)) continue;
        let label = getButtonLabel.call(self, node);

        let reference;
        if (node.hasAttribute("id")) {
            reference = node.getAttribute("id");
        }

        let disabled;
        if (node.hasAttribute("disabled") || node.disabled === true) {
            disabled = true;
        }

        if (!reference) {
            reference = new ID("tab").toString();
            node.setAttribute("id", reference);
        }

        if (node.hasAttribute(`${ATTRIBUTE_PREFIX}button-icon`)) {
            label = `<span part="label">${label}</span><img part="icon" src="${node.getAttribute(
                `${ATTRIBUTE_PREFIX}button-icon`,
            )}">`;
        }

        let remove = false;
        if (node.hasAttribute(`${ATTRIBUTE_PREFIX}removable`)) {
            remove = true;
        }

        if (node.matches(".active") === true && disabled !== true) {
            node.classList.remove("active");
            activeReference = reference;
        }

        const state = "";
        const classes = dimensionsCalculated ? "" : "invisible";

        buttons.push({
            reference,
            label,
            state,
            class: classes,
            disabled,
            remove,
        });

        attachTabMutationObserver.call(self, node);
    }

    self.setOption("buttons.standard", clone(buttons));
    self.setOption("buttons.popper", []);
    self.setOption("marker", random());

    return adjustButtonVisibility.call(self).then(() => {
        if (activeReference) {
            return new Processing(() => {
                const button = self.shadowRoot.querySelector(`[${ATTRIBUTE_PREFIX}tab-reference="${activeReference}"]`);
                if (button instanceof HTMLButtonElement && button.disabled !== true) {
                    show.call(self, button);
                }
            }).run();
        }

        return;
    });
}

function checkAndRearrangeButtons() {
    const self = this;

    if (self[dimensionsSymbol].getVia("data.calculated", false) !== true) {
        calculateNavigationButtonsDimensions.call(self);
    }

    rearrangeButtons.call(self);
}

/**
 * @private
 * @return {Promise<unknown>}
 */
function adjustButtonVisibility() {
    const self = this;

    return new Promise((resolve) => {
        const observer = new MutationObserver(function (mutations) {
            const defCount = self.getOption("buttons.standard").length;
            const domCount = self[navElementSymbol].querySelectorAll('button[data-monster-role="button"]').length;

            // in drawing
            if (defCount !== domCount) return;

            observer.disconnect();

            checkAndRearrangeButtons.call(self);

            resolve();
        });

        observer.observe(self[navElementSymbol], {
            attributes: true,
        });
    });
}

/**
 * @private
 * @param node
 * @return {number}
 */
function getDimValue(value) {
    if ([undefined, null].indexOf(value) !== -1) {
        return 0;
    }

    const valueAsInt = parseInt(value, 10);

    if (isNaN(valueAsInt)) {
        return 0;
    }

    return valueAsInt;
}

/**
 * @private
 * @param {HTMLElement} node
 * @return {number}
 */
function calcBoxWidth(node) {
    let dim = getGlobal("window").getComputedStyle(node);
    let bounding = node.getBoundingClientRect();

    return (
        getDimValue(dim["border-left-width"]) +
        getDimValue(dim["padding-left"]) +
        getDimValue(dim["margin-left"]) +
        getDimValue(bounding["width"]) +
        getDimValue(dim["border-right-width"]) +
        getDimValue(dim["margin-right"]) +
        getDimValue(dim["padding-left"])
    );
}

/**
 * @private
 * @return {Object}
 */
function rearrangeButtons() {
    const self = this;

    const standardButtons = [];
    const popperButtons = [];

    let sum = 0;
    const space = self[dimensionsSymbol].getVia("data.space");

    const buttons = self.getOption("buttons.standard");
    for (const [, button] of buttons.entries()) {
        const ref = button?.reference;

        sum += self[dimensionsSymbol].getVia(`data.button.${ref}`);

        if (sum > space) {
            popperButtons.push(clone(button));
        } else {
            standardButtons.push(clone(button));
        }
    }

    self.setOption("buttons.standard", standardButtons);
    self.setOption("buttons.popper", popperButtons);

    if (self[switchElementSymbol]) {
        if (popperButtons.length > 0) {
            self[switchElementSymbol].classList.remove("hidden");
        } else {
            self[switchElementSymbol].classList.add("hidden");
        }
    }
}

/**
 * @private
 * @return {Object}
 */
function calculateNavigationButtonsDimensions() {
    const self = this;

    const width = self[navElementSymbol].getBoundingClientRect().width;

    let startEndWidth = 0;

    getSlottedElements.call(self, undefined, "start").forEach((node) => {
        startEndWidth += calcBoxWidth.call(self, node);
    });

    getSlottedElements.call(self, undefined, "end").forEach((node) => {
        startEndWidth += calcBoxWidth.call(self, node);
    });

    self[dimensionsSymbol].setVia("data.space", width - startEndWidth - 2);
    self[dimensionsSymbol].setVia("data.visible", !(width === 0));

    const buttons = self.getOption("buttons.standard").concat(self.getOption("buttons.popper"));

    for (const [i, button] of buttons.entries()) {
        const ref = button?.reference;
        const element = self[navElementSymbol].querySelector(`:scope > [${ATTRIBUTE_PREFIX}tab-reference="${ref}"]`);
        if (!(element instanceof HTMLButtonElement)) continue;

        self[dimensionsSymbol].setVia(`data.button.${ref}`, calcBoxWidth.call(self, element));
        button["class"] = new TokenList(button["class"]).remove("invisible").toString();
    }

    const slots = self[controlElementSymbol].querySelectorAll(
        `nav[${ATTRIBUTE_PREFIX}role=nav] > slot.invisible, slot[${ATTRIBUTE_PREFIX}role=slot].invisible`,
    );
    for (const [, slot] of slots.entries()) {
        slot.classList.remove("invisible");
    }

    self[dimensionsSymbol].setVia("data.calculated", true);
    self.setOption("buttons.standard", clone(buttons));
}

/**
 * @private
 * @param {HTMLElement} node
 * @return {string}
 */
function getButtonLabel(node) {
    const self = this;

    let label;
    let setLabel = false;
    if (node.hasAttribute(ATTRIBUTE_BUTTON_LABEL)) {
        label = node.getAttribute(ATTRIBUTE_BUTTON_LABEL);
    } else {
        label = node.innerText;
        setLabel = true;
    }

    if (!isString(label)) {
        label = "";
    }

    label = label.trim();

    if (label === "") {
        label = self.getOption("labels.new-tab-label", "New Tab");
    }

    if (label.length > 100) {
        label = `${label.substring(0, 99)}…`;
    }

    if (setLabel === true) {
        node.setAttribute(ATTRIBUTE_BUTTON_LABEL, label);
    }

    return label;
}

/**
 * @private
 * @return {string}
 */
function getTemplate() {
    // language=HTML
    return `
        <template id="buttons">
            <button part="button"
                    data-monster-role="button"
                    data-monster-attributes="
                    data-monster-state path:buttons.state,
                    disabled path:buttons.disabled | if:true,                    
                    data-monster-tab-reference path:buttons.reference"><span
                    data-monster-replace="path:buttons.label"></span><span part="remove-tab"
                                                                           data-monster-attributes="class path:buttons.remove | ?:remove-tab:hidden "
                                                                           data-monster-role="remove-tab"
                                                                           tabindex="-1"></span></button>
        </template>
        <div data-monster-role="control" part="control">
            <nav data-monster-role="nav" part="nav"
                 data-monster-attributes="data-monster-marker path:marker"
                 data-monster-insert="buttons path:buttons.standard">
                <slot name="start" class="invisible"></slot>
                <div data-monster-role="popper" part="popper" tabindex="-1" class="monster-color-primary-1">
                    <div data-popper-arrow></div>


                    <div part="popper-nav" data-monster-role="popper-nav"
                         data-monster-insert="buttons path:buttons.popper"
                         tabindex="-1"></div>
                </div>
                <slot name="popper" class="invisible"></slot>
                <slot name="end" class="invisible"></slot>
            </nav>
            <slot data-monster-role="slot" class="invisible"></slot>
        </div>`;
}

registerCustomElement(Tabs);
