/**
 * Copyright schukai GmbH and contributors 2023. All Rights Reserved.
 * Node module: @schukai/component-form
 * This file is licensed under the AGPLv3 License.
 * License text available at https://www.gnu.org/licenses/agpl-3.0.en.html
 */

import { buildTree } from "@schukai/monster/source/data/buildtree.mjs";
import { findClosestByAttribute } from "@schukai/monster/source/dom/attributes.mjs";
import { ATTRIBUTE_ROLE, ATTRIBUTE_UPDATER_INSERT_REFERENCE } from "@schukai/monster/source/dom/constants.mjs";
import { instanceSymbol } from "@schukai/monster/source/constants.mjs";
import { assembleMethodSymbol, registerCustomElement } from "@schukai/monster/source/dom/customelement.mjs";
import { findTargetElementFromEvent, fireEvent } from "@schukai/monster/source/dom/events.mjs";
import { Formatter } from "@schukai/monster/source/text/formatter.mjs";
import { isString } from "@schukai/monster/source/types/is.mjs";
import { Node } from "@schukai/monster/source/types/node.mjs";
import { NodeRecursiveIterator } from "@schukai/monster/source/types/noderecursiveiterator.mjs";
import { validateInstance } from "@schukai/monster/source/types/validate.mjs";
import { ATTRIBUTE_FORM_URL, ATTRIBUTE_INTEND } from "./constants.mjs";
import { Select } from "./select.mjs";
import { SelectStyleSheet } from "./stylesheet/select.mjs";
import { TreeSelectStyleSheet } from "./stylesheet/tree-select.mjs";

export { TreeSelect, formatHierarchicalSelection };

/**
 * @private
 * @type {symbol}
 */
const internalNodesSymbol = Symbol("internalNodes");

/**
 * @private
 * @type {symbol}
 */
const keyEventHandler = Symbol("keyEventHandler");

/**
 * This CustomControl creates a select element with a variety of options.
 * It supports filtering, local and remote, multiple selection and has a
 * template system for displaying the options.
 *
 * <img src="./images/treeselect.png">
 *
 * Dependencies: the system uses functions of the [monsterjs](https://monsterjs.org/) library
 * as well as [pooperjs](https://popper.js.org/docs/v2/).
 *
 * You can create this control either by specifying the HTML tag `<monster-select />` directly in the HTML or using
 * Javascript via the `document.createElement('monster-select');` method.
 *
 * ```html
 * <monster-select></monster-select>
 * ```
 *
 * Or you can create this CustomControl directly in Javascript:
 *
 * ```js
 * import {TreeSelect} from '@schukai/component-form/source/treeselect.js';
 * document.createElement('monster-tree-select');
 * ```
 *
 * @externalExample ../example/tree-select.mjs
 * @startuml treeselect.png
 * skinparam monochrome true
 * skinparam shadowing false
 * HTMLElement <|-- CustomElement
 * CustomElement <|-- CustomControl
 * CustomControl <|-- Select
 * Select <|-- TreeSelect
 * @enduml
 *
 * @since 1.9.0
 * @copyright schukai GmbH
 * @memberOf Monster.Components.Form
 * @summary A highly configurable select control
 */
class TreeSelect extends Select {
    /**
     * This method is called by the `instanceof` operator.
     * @returns {symbol}
     * @since 2.1.0
     */
    static get [instanceSymbol]() {
        return Symbol.for("@schukai/component-form/tree-select@@instance");
    }

    /**
     * To set the options via the html tag the attribute `data-monster-options` must be used.
     * @see {@link https://monsterjs.org/en/doc/#configurate-a-monster-control}
     *
     * The individual configuration values can be found in the table.
     *
     * @extends Monster.Components.Form.Select
     * @property {String} mapping.rootReferences=['0', undefined, null]
     * @property {String} mapping.idTemplate=id
     * @property {String} mapping.parentTemplate=parent
     * @property {String} mapping.selection
     * @property {Object} formatter
     * @property {String} formatter.separator=" / "
     */
    get defaults() {
        return Object.assign(
            {},
            super.defaults,
            {
                mapping: {
                    rootReferences: ["0", undefined, null],
                    idTemplate: "id",
                    parentTemplate: "parent",
                },
                formatter: {
                    selection: formatHierarchicalSelection,
                    separator: " / ",
                },
                templates: {
                    main: getTemplate(),
                },
            },
            initOptionsFromArguments.call(this),
        );
    }

    /**
     *
     * @return {string}
     */
    static getTag() {
        return "monster-tree-select";
    }

    /**
     *
     * @return {CSSStyleSheet[]}
     */
    static getCSSStyleSheet() {
        return [SelectStyleSheet, TreeSelectStyleSheet];
    }

    /**
     * Import Select Options from dataset
     * Not to be confused with the control defaults/options
     *
     * @since 0.16.0
     * @param {array|object|Map|Set} data
     * @return {Select}
     * @throws {Error} map is not iterable
     */
    importOptions(data) {
        const self = this;

        self[internalNodesSymbol] = new Map();

        const mappingOptions = self.getOption("mapping", {});

        const filter = mappingOptions?.["filter"];
        const rootReferences = mappingOptions?.["rootReferences"];

        const id = self.getOption("mapping.idTemplate", "id");
        const parentID = self.getOption("mapping.parentTemplate", "parent");

        const selector = mappingOptions?.["selector"];

        let nodes = buildTree(data, selector, id, parentID, {
            filter,
            rootReferences,
        });

        let options = [];
        for (const node of nodes) {
            const iterator = new NodeRecursiveIterator(node);
            for (const n of iterator) {
                const formattedValues = formatKeyLabel.call(self, n);

                const label = formattedValues.label;
                const value = formattedValues.value;
                const intend = n.level;

                let visibility = intend > 0 ? "hidden" : "visible";
                let state = "close";

                self[internalNodesSymbol].set(value, n);

                options.push({
                    value,
                    label,
                    intend,
                    state,
                    visibility,
                    ["has-children"]: n.hasChildNodes(),
                });
            }
        }

        self.setOption("options", options);
        return self;
    }

    /**
     *
     * @return {Monster.Components.Form.Select}
     */
    [assembleMethodSymbol]() {
        super[assembleMethodSymbol]();
        initEventhandler.call(this);
    }
}

/**
 * @private
 * @param event
 */
function handleOptionKeyboardEvents(event) {
    const self = this;

    switch (event?.["code"]) {
        case "ArrowLeft":
            closeOrOpenCurrentOption.call(self, event, "close");
            event.preventDefault();
            break;
        case "ArrowRight":
            closeOrOpenCurrentOption.call(self, event, "open");
            event.preventDefault();
            break;
    }
}

/**
 * @private
 * @param {event} event
 */
function closeOrOpenCurrentOption(event, mode) {
    validateInstance(event, Event);

    if (typeof event.composedPath !== "function") {
        throw new Error("unsupported event");
    }

    const path = event.composedPath();
    const optionNode = path.shift();

    const state = optionNode.getAttribute("data-monster-state");
    if (state !== mode) {
        const handler = optionNode.querySelector("[data-monster-role=folder-handler]");
        if (handler instanceof HTMLElement) {
            fireEvent(handler, "click");
        }
    }
}

/**
 *
 * @param {Node} node
 * @return {array<label, value>}
 * @memberOf Monster.Components.Form
 * @private
 */
function formatKeyLabel(node) {
    const self = this;
    validateInstance(node, Node);

    const label = new Formatter(node.value).format(self.getOption("mapping.labelTemplate", ""));
    const value = new Formatter(node.value).format(self.getOption("mapping.valueTemplate", ""));

    return {
        value,
        label,
    };
}

/**
 * @private
 * @param {string} value
 * @return {Array}
 */
function buildTreeLabels(value) {
    const self = this;
    let node = self[internalNodesSymbol].get(value);
    if (node === undefined) {
        node = self[internalNodesSymbol].get(parseInt(value));
    }

    const parts = [];

    if (node instanceof Node) {
        let ptr = node;
        while (ptr) {
            const formattedValues = formatKeyLabel.call(self, ptr);
            parts.unshift(formattedValues.label);
            ptr = ptr.parent;
        }
    }

    return parts;
}

/**
 * This formatter can format a label hierarchically.
 * The option `formatter.separator` determines the separator.
 *
 * ```
 * a / b / c
 * ```
 *
 * This function can be passed as argument of the option `formatter.selection:`.
 *
 * @since 1.9.0
 * @param {*} value
 * @return {string}
 * @memberOf Monster.Components.Form
 */
function formatHierarchicalSelection(value) {
    const self = this;
    return buildTreeLabels.call(self, value).join(self.getOption("formatter.separator", " / "));
}

/**
 * @private
 * @type {symbol}
 */
const openOptionEventHandler = Symbol("openOptionEventHandler");

/**
 * @private
 * @throws {Error} no shadow-root is defined
 */
function initEventhandler() {
    const self = this;

    if (!self.shadowRoot) {
        throw new Error("no shadow-root is defined");
    }

    self[openOptionEventHandler] = (event) => {
        const element = findTargetElementFromEvent(event, ATTRIBUTE_ROLE, "folder-handler");
        if (!(element instanceof HTMLElement)) {
            return;
        }

        let container = findClosestByAttribute(element, ATTRIBUTE_ROLE, "option");
        const index = container.getAttribute(ATTRIBUTE_UPDATER_INSERT_REFERENCE).split("-").pop();

        const currentState = self.getOption(`options.${index}.state`);

        const newState = currentState === "close" ? "open" : "close";
        self.setOption(`options.${index}.state`, newState);

        const newVisibility = newState === "open" ? "visible" : "hidden";

        if (container.hasAttribute(ATTRIBUTE_INTEND)) {
            const intend = container.getAttribute(ATTRIBUTE_INTEND);

            let ref = container.nextElementSibling;
            const childIntend = parseInt(intend) + 1;

            const cmp = (a, b) => {
                if (newState === "open") {
                    return a === b;
                }

                return a >= b;
            };

            while (
                ref?.hasAttribute(ATTRIBUTE_INTEND) &&
                cmp(parseInt(ref.getAttribute(ATTRIBUTE_INTEND)), childIntend)
            ) {
                const refIndex = ref.getAttribute(ATTRIBUTE_UPDATER_INSERT_REFERENCE).split("-").pop();
                self.setOption(`options.${refIndex}.visibility`, newVisibility);

                if (newState === "close") {
                    self.setOption(`options.${refIndex}.state`, "close");
                }

                ref = ref.nextElementSibling;
            }
        }
    };

    self[keyEventHandler] = (event) => {
        const path = event.composedPath();
        let element = path?.[0];

        let role;

        if (element instanceof HTMLElement) {
            if (element.hasAttribute(ATTRIBUTE_ROLE)) {
                role = element.getAttribute(ATTRIBUTE_ROLE);
            } else if (element === this) {
                show.call(this);
                focusFilter.call(self);
            } else {
                let e = element.closest(`[${ATTRIBUTE_ROLE}]`);
                if (e instanceof HTMLElement && e.hasAttribute()) {
                    role = e.getAttribute(ATTRIBUTE_ROLE);
                }
            }
        } else {
            return;
        }

        switch (role) {
            case "option-label":
            case "option-control":
            case "option":
                handleOptionKeyboardEvents.call(self, event);
                break;
        }
    };

    self.shadowRoot.addEventListener("keydown", self[keyEventHandler]);
    self.shadowRoot.addEventListener("click", self[openOptionEventHandler]);
}

/**
 * This attribute can be used to pass a URL to this select.
 *
 * ```html
 * <monster-select data-monster-url="https://example.com/"></monster-select>
 * ```
 *
 * @private
 * @return {object}
 */
function initOptionsFromArguments() {
    const self = this;
    let options = {};

    let url = self.getAttribute(ATTRIBUTE_FORM_URL);
    if (isString(url)) {
        options["url"] = new URL(url).toString();
    }

    return options;
}

/**
 * @private
 * @return {string}
 */
function getTemplate() {
    // language=HTML
    return `
        <template id="options">
            <div data-monster-role="option"
                 tabindex="-1"
                 data-monster-attributes="
                 data-monster-intend path:options.intend, 
                 data-monster-state path:options.state, 
                 data-monster-visibility path:options.visibility, 
                 data-monster-filtered path:options.filtered,
                 data-monster-has-children path:options.has-children">
                <div data-monster-role="folder-handler"></div>
                <label part="option" role="option">
                    <input data-monster-role="option-control"
                           data-monster-attributes="
            type path:type,
            role path:role,
            value path:options.value, 
            name path:name, 
            part path:type | prefix:option- | suffix: form
            " tabindex="-1">
                    <span data-monster-replace="path:options | index:label" part="option-label"></span>
                </label>
            </div>
        </template>

        <template id="selection">
            <div data-monster-role="badge"
                 part="badge"
                 data-monster-attributes="
                 data-monster-value path:selection | index:value, 
                 class path:classes | index:badge, 
        part path:type | suffix:-option | prefix: form-" tabindex="-1">
                <div data-monster-replace="path:selection | index:label" part="badge-label"
                     data-monster-role="badge-label"></div>
                <div part="remove-badge" data-monster-select-this
                     data-monster-attributes="class path:features.clear | ?::hidden "
                     data-monster-role="remove-badge" tabindex="-1"></div>
            </div>
        </template>

        <slot class="hidden"></slot>

        <div data-monster-role="control" part="control" tabindex="0">
            <div data-monster-role="container">
                \${selected}
            </div>

            <div data-monster-role="popper" part="popper" tabindex="-1" class="monster-color-primary-1">
                <div class="option-filter-control" role="search">
                    <input type="text" role="searchbox"
                           part="popper-filter" name="popper-filter"
                           data-monster-role="filter"
                           autocomplete="off"
                           tabindex="0">
                </div>
                <div part="content" class="flex" data-monster-replace="path:content">
                    <div part="options" data-monster-role="options" data-monster-insert="options path:options"
                         tabindex="-1"></div>
                </div>
                <div part="no-options" data-monster-role="no-options"
                     data-monster-replace="path:messages.emptyOptions"></div>
            </div>
            <div part="status-or-remove-badges" data-monster-role="status-or-remove-badges"
                 data-monster-attributes="class path:classes.statusOrRemoveBadge | suffix:\\ status-or-remove-badges"></div>
        </div>
    `;
}

registerCustomElement(TreeSelect);
